import os
import threading
from datetime import datetime
from .Singleton import Singleton
from .KowanasTime import KowanasTime

class Log(metaclass=Singleton):
    __lock = threading.Lock()
    __logfile = None
    def __init__(self, logfile = None):
        self.__logfile = logfile
        if self.__logfile != None:
            print ('log' + self.__logfile)
            if os.path.exists(self.__logfile) == False:
                with open(self.__logfile, 'w') as f:
                    f.write('created at '+self.__getNow())

    def __getNow(self):
        return KowanasTime.getKST().strftime('%Y-%m-%d %H:%M:%S')

    def d(self, *log):
        strbuf = ''
        for l in log:
            strbuf += str(l)

        print (strbuf)
#        thread = threading.Thread(target=self.write, args=(log),)
#        thread.start()
        if self.__logfile != None: 
            self.__write(self.__getNow()+': '+strbuf+'\n')

    def __write(self, log):
        self.__lock.acquire()
        with open(self.__logfile, 'a') as f:
            f.write(log)
        self.__lock.release()


from .SingletonClass import SingletonClass
from threading import Thread

def run(name, result):
    singleton = SingletonClass(name)
    result[0] = str(singleton)

def test_Singleton():
    resultA = ['A']
    resultB = ['B']
    processa = Thread(target=run, args=('ACLASS', resultA))
    processb = Thread(target=run, args=('BCLASS', resultB))
    processa.start()
    processb.start()
    assert(resultA[0] == resultB[0])


from .Singleton import Singleton

class Config(metaclass=Singleton):
    configs = None

    def __init__(self, file='/home/ubuntu/workspace/.config'):
        with open(file) as f:
            self.configs = self._parse(f.read())

    def _parse(self, data):
        configs = {}
        lines = data.split('\n')
        for line in lines:
            if '=' not in line: continue
            line = line.replace(' ', '')
            seperated = line.split('=')
            key = seperated[0]
            value = seperated[1]
            configs[key] = value
        return configs

    def get(self, key):
        return self.configs[key]



from .DBConnection import DBConnection
from .DBModel import DBModel
from .Config import Config
from .Singleton import Singleton
from .Service import Service
from .Task import Task
from .ApiResult import ApiResult
from .Log import Log
from .KowanasTime import KowanasTime
from .ApiHandler import ApiHandler
from .DBField import DBField
from .DBFieldUID import DBFieldUID
from .Scheduler import Scheduler


import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
from util import Config

def test_Config():
    config = Config(file='./data/config_example')
    assert config.get('DB_HOST') == '123.234.111.1'
    assert config.get('DB_USER') == 'test'
    assert config.get('DB_PASSWORD') == 'testpwd123'


import json

class ApiResult:
    result = {}

    def __init__(self, success = False):
        self.result['success'] = success

    def toJson(self):
        return json.dumps(self.result)

    def update(self, key, data):
        self.result[key] = data


from datetime import datetime
from datetime import timedelta

class Task:
    def __init__(self, name, handler, args, schedule=None):
        self.name = name
        self.__handler = handler
        self.__args = args
        self.schedule = schedule

    def run(self):
        self.__handler(self.__args)


from datetime import datetime
from datetime import timedelta

class KowanasTime:
    @classmethod
    def getKST(self, setZeroSec = False, setZeroMin = False, setZeroHour = False):
        now = datetime.now()+timedelta(hours=9)
        if setZeroHour == True:
            now -= timedelta(hours=now.hour, minutes=now.minute, seconds=now.second)
        elif setZeroMin == True:
            now -= timedelta(minutes=now.minute, seconds=now.second)
        elif setZeroSec == True:
            now -= timedelta(seconds=now.second)
        return now

    @classmethod
    def today(self):
        now = datetime.now()+timedelta(hours=9)
        return now.strftime('%Y%m%d')

    @classmethod
    def getTimeStamp(self):
        return datetime.now().timestamp


import json
from abc import ABCMeta
from .ApiResult import ApiResult

class ApiHandler(metaclass=ABCMeta):
    def handle(self, api, args):
        if args['type'] == 'read':
            return self._handleRead(api, args)
        else:
            return self._handle(api, args)

    def handlePost(self, api, args):
        if args['type'] == 'create':
            return self._handleCreate(api, args)
        else:
            return self._handlePost(api, args)

    def _handle(self, api, args):
        pass

    def _handlePost(self, api, args):
        pass

    def _handleCreate(self, api, args):
        pass

    def _handleUpdate(self, api, args):
        pass

    def _handleRead(self, api, args):
        pass

    def _handleDelete(self, api, args):
        pass

    def _createResult(self, success = True, data = []):
        result = ApiResult(success = success)
        result.update('data', data)
        return result



from abc import ABCMeta

class DBModel(metaclass=ABCMeta):
    __dbConnection = None
    __tablename = None
    __fields = None
    __foreignkeys = None
    _records = {}

    AUTO_INCREMENT = -1
    NO_RECORD = "NO_RECORD"

    String = "VARCHAR"
    Int = "INT"
    Float = "FLOAT"
    Long = "BIGINT"

    def __init__(self, dbConnection, tablename, fields=None, foreignkeys=[], debug=False):
        self.__dbConnection = dbConnection
        self.__tableName = tablename
        self.__foreignkeys = foreignkeys
        self.__debug = debug
        if fields != None:
            self.__fields = fields
            self.__createTable()

    def __createTable(self):
        sql = 'CREATE TABLE if not exists '+self.__tableName+' ('
        for field in self.__fields:
            sql += field.key + ' ' + field.type + '(' + str(field.length) + ') NOT NULL ' + field.autoIncrement + ', '
        sql += 'PRIMARY KEY ('+ self.__fields[0].key +')'
        for foreignkey in self.__foreignkeys:
            sql += ', FOREIGN KEY ('+foreignkey[0]+') REFERENCES '+foreignkey[1]
        sql += ') ENGINE=InnoDB, charset=utf8;'
        return self.__query(sql)

    def __query(self, sql):
        if self.__debug == True:
            print (sql)
        try:
            self.__dbConnection.cursor.execute(sql)
        except Exception as e:
            print (str(e))
            return False
        return True

    def __fetch(self):
        return self.__dbConnection.cursor.fetchall()    

    def commit(self):
        try:
            self.__dbConnection.commit()
        except Exception as e:
            print (e)
            return False
        return True

    def __columnsToString(self, columns):
        sqlcolumns = ''
        if columns != None:
            sqlcolumns = '('+columns+')'
        return sqlcolumns    

    def __whereToString(self, where):
        sqlwhere = ''
        keys= list(where.keys())
        values = list(where.values())
        for key, value in zip(keys, values):
            sqlwhere = sqlwhere + key + '='
            if type(value) is not str:
                value = str(value)
            sqlwhere = sqlwhere + value

    def _insert(self, values, columns = None):
        sqlcolumns = self.__columnsToString(columns)
        sqlvalues = '('+values+')'
        sql = 'INSERT INTO '+self.__tableName+' '+sqlcolumns+' VALUES '+sqlvalues+';'
        return self.__query(sql)

    def _update(self, values, where = None):
        sql = 'UPDATE '+self.__tableName+' SET '+values+' WHERE '+where+';'
#        sql = 'REPLACE INTO '+self.__tableName+' SET '+values+';'
        return self.__query(sql)

    def _select(self, where = None):
        sql = 'SELECT * FROM '+self.__tableName
        if where != None: sql += ' WHERE '+where+';'
        else: sql += ';'
        if self.__query(sql) == False: return []
        return self.__fetch()

    def _delete(self, where):
        sql = 'DELETE FROM '+self.__tableName+' WHERE '+where+';'
        if self.__query(sql) == False: return False
        return True

    # clear table
    def clear(self):
        sql = 'truncate table '+self.__tableName+';'
        self.__query(sql)
        result = self.__dbConnection.cursor.fetchall()
        return result

    def __getStringByFieldType(self, data, field):
        if field.type == DBModel.String:
            return '\"'+data+'\"'
        if field.type == DBModel.Int or field.type == DBModel.Float or field.type == DBModel.Long:
            return str(data)

    # get query string for query
    def _getInsertQuery(self, data):
        if data[0] == DBModel.AUTO_INCREMENT: sql = 'NULL'
        else: sql = self.__getStringByFieldType(data[0], self.__fields[0])
        for index in range(1, len(data)):
            sql += ', ' + self.__getStringByFieldType(data[index], self.__fields[index])
        return sql

    def _getUpdateQuery(self, data):
        sql = ''
        for field, value in zip(self.__fields[1:], data[1:]):
            if field.type == DBModel.String:
                sql += field.key + '=\'' + str(value) + '\','
            else:
                sql += field.key + '=' + str(value) + ','
        return sql[:-1]

    # compare between 2 data
    def equal(self, a, b):
        pass

    # verify data
    def _verify(self, data):
        return True

    # get data from row
    def _fromRow(self, row):
        if len(row) > 0:
            return [row[field.key] for field in self.__fields]
        else: return []

    # convert data to json
    def toJson(self, data):
        return json.dumps(_records)

    # add data to table
    def add(self, data = None, columns = None):
        if self._verify(data) == False: return False
        if self._insert(self._getInsertQuery(data), columns = columns) == False: 
            return False
        return True

    # update record in table
    def update(self, uid, data):
        if self._verify(data) == False: return False
        where = 'uid = \''+str(uid)+'\''
        if self._update(self._getUpdateQuery(data), where = where) == False: 
            return False
        return True

    # read data from table
    def read(self, uid):
        where = 'uid = \''+str(uid)+'\''
        rows = self._select(where)
        if len(rows) <= 0: return False
        return self._fromRow(rows[0])

    def cache(self):
        self._records = self.readAll()

    # read all data from table
    def readAll(self):
        rows = self._select()
        records = []
        for row in rows:
            record = self._fromRow(row)
            records.append(record)
#            records[record[0]] = record
        return records

    def readAllToDict(self):
        rows = self._select()
        records = {}
        for row in rows:
            record = self._fromRow(row)
            records[record[0]] = record
        return records

    # delete record in table
    def delete(self, uid):
        where = 'uid = '+str(uid)
        return self._delete(where)
